/*
 * mvc.c
 *
 * Copyright (C) Merlin Ma. 2000-2004. 
 * EMail: merlin@turbolinux.com.cn 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <libgen.h>
#include <errno.h>
#include <getopt.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <linux/videodev.h>
#ifdef HAVE_JPEG
#include <jpeglib.h>
#endif
#ifdef HAVE_PNG
#include <png.h>
#endif

#include "font.h"
#include "config.h"

typedef struct myvideo {
    char *device;
    int devfd;
    int width;
    int height;
    int depth;
    int palette;
    int input;
    int maxinputs;
    int norm;
    unsigned int brightness;
    unsigned int contrast;
    char *buf;
    int mmaped;
    int mbufsize;
} MyVID;

/* global variable */
static int verbose = DEF_VERBOSE;
static int quitflag = 0;
static int daemon_flag = 0;
static int tstp_flag = 0;
static int usr1_flag = 0;
static char *logfile = "/dev/null";

/* SIGINT (Ctrl-C) signal handler */
void sig_quit(int signal)
{
    printf("Got SIGINT, Terminating programm...\n");
    quitflag = 1;
}

/* SIGTSTP (Ctrl-Z) signal handler */
void sig_tstp(int signal)
{
    printf("Got SIGTSTP, switching to daemon mode.\n");
    tstp_flag = 1;
}

/* USER1 signal handler */
void sig_usr1(int signal)
{
    printf("Got SIGUSR1, trigger a capture.\n");
    usr1_flag = 1;
}

void block_sigs(void)
{
    sigset_t masked_signals;
    sigemptyset(&masked_signals);
    sigaddset(&masked_signals, SIGUSR1);
    sigaddset(&masked_signals, SIGINT);
    sigaddset(&masked_signals, SIGTSTP);
    sigprocmask(SIG_BLOCK, &masked_signals, NULL);
}

void unblock_sigs(void)
{
    sigset_t masked_signals;
    sigemptyset(&masked_signals);
    sigaddset(&masked_signals, SIGUSR1);
    sigaddset(&masked_signals, SIGINT);
    sigaddset(&masked_signals, SIGTSTP);
    sigprocmask(SIG_UNBLOCK, &masked_signals, NULL);
}

void usage(void)
{
    fprintf(stderr,
	    "mvc version " VERSION " - motion detection video capture program \n\n"
	    "Usage: mvc <options>\n\n"
	    " -B picture brightness value.\n"
	    " -C picture contrast value.\n"
	    " -D run in daemon mode.\n"
	    " -O Base directory for output images. default is current dir.\n"
	    " -o Don't create sub-directories for output.\n"
	    " -t Print timestamp on image.\n"
	    " -d <device> video device (default:" VIDEO_DEV ").\n"
	    " -e <program> excute a external program while motion detected.\n"
	    " -f {jpeg|png} set output file format. default is jpeg.\n"
	    " -q <quality>               jpeg quality setting.\n"
	    " -h print this help.\n"
	    " -i <channel> which input channel to use, default is 1, or 0 if only 1 input.\n"
	    " -m use mono color (grayscale) for capture.\n"
	    " -n {pal|ntsc|secam|auto}    video norm to use, default: PAL\n"
	    " -s XxY size of the output image, example: -s 640x480\n"
	    " -a HxV block division number in Horizantal and Vertical. Ex: -a 16x8\n"
	    " -b <diff_block_limit>      threshold of hot block number.\n"
	    " -c <diff_color_limit>      threshold of block color difference.\n"
	    " -w delay time for each capture(seconds), example: -w 0.5\n"
	    " -l logfile used in daemon mode. default is null.\n"
	    " -v increase verbose level for more details, can multiple applied.\n\n");
}

void daemonizenow(void)
{
    pid_t pid;
    pid_t sid;
    int fd;

    pid = fork();
    if (pid == -1)
	printf("Failed to fork()\n\n");

    if (pid) {
	if (verbose)
	    printf("Daemon born, PID %d, detaching...\n\n", pid);
	exit(0);
    }

    fd = open(logfile, O_RDWR | O_CREAT | O_APPEND, 00600);
    if (fd != -1) {
	if (fd != 0)
	    dup2(fd, 0);
	if (fd != 1)
	    dup2(fd, 1);
	if (fd != 2)
	    dup2(fd, 2);
	if (fd > 2)
	    close(fd);
    }
    else {
	printf("Output redirection error.\n");
    }

    sid = setsid();
    if (sid == -1)
	abort();
    return;
}

/*
 * Turn a YUV4:2:0 block into an RGB block
 * This code is taken from v4l.c of w3cam by Rasca, Berlin 1998-2001.
 *
 * Video4Linux seems to use the blue, green, red channel
 * order convention-- rgb[0] is blue, rgb[1] is green, rgb[2] is red.
 *
 * Color space conversion coefficients taken from the excellent
 * http://www.inforamp.net/~poynton/ColorFAQ.html
 * In his terminology, this is a CCIR 601.1 YCbCr -> RGB.
 * Y values are given for all 4 pixels, but the U (Pb)
 * and V (Pr) are assumed constant over the 2x2 block.
 *
 * To avoid floating point arithmetic, the color conversion
 * coefficients are scaled into 16.16 fixed-point integers.
 * They were determined as follows:
 *
 *	double brightness = 1.0;  (0->black; 1->full scale) 
 *	double saturation = 1.0;  (0->greyscale; 1->full color)
 *	double fixScale = brightness * 256 * 256;
 *	int rvScale = (int)(1.402 * saturation * fixScale);
 *	int guScale = (int)(-0.344136 * saturation * fixScale);
 *	int gvScale = (int)(-0.714136 * saturation * fixScale);
 *	int buScale = (int)(1.772 * saturation * fixScale);
 *	int yScale = (int)(fixScale);	
 */

/* LIMIT: convert a 16.16 fixed-point value to a byte, with clipping. */
#define LIMIT(x) ((x)>0xffffff?0xff: ((x)<=0xffff?0:((x)>>16)))

static inline void v4l_copy_420_block (int yTL, int yTR, int yBL, int yBR, int u, int v, 
	int rowPixels, unsigned char * rgb, int bits)
{
	const int rvScale = 91881;
	const int guScale = -22553;
	const int gvScale = -46801;
	const int buScale = 116129;
	const int yScale  = 65536;
	int r, g, b;

	g = guScale * u + gvScale * v;
	r = rvScale * v;
	b = buScale * u;

	yTL *= yScale; yTR *= yScale;
	yBL *= yScale; yBR *= yScale;

	if (bits == 24) {
		/* Write out top two pixels */
		rgb[0] = LIMIT(b+yTL); rgb[1] = LIMIT(g+yTL); rgb[2] = LIMIT(r+yTL);
		rgb[3] = LIMIT(b+yTR); rgb[4] = LIMIT(g+yTR); rgb[5] = LIMIT(r+yTR);

		/* Skip down to next line to write out bottom two pixels */
		rgb += 3 * rowPixels;
		rgb[0] = LIMIT(b+yBL); rgb[1] = LIMIT(g+yBL); rgb[2] = LIMIT(r+yBL);
		rgb[3] = LIMIT(b+yBR); rgb[4] = LIMIT(g+yBR); rgb[5] = LIMIT(r+yBR);
	} else if (bits == 16) {
		/* Write out top two pixels */
		rgb[0] = ((LIMIT(b+yTL) >> 3) & 0x1F) | ((LIMIT(g+yTL) << 3) & 0xE0);
		rgb[1] = ((LIMIT(g+yTL) >> 5) & 0x07) | (LIMIT(r+yTL) & 0xF8);

		rgb[2] = ((LIMIT(b+yTR) >> 3) & 0x1F) | ((LIMIT(g+yTR) << 3) & 0xE0);
		rgb[3] = ((LIMIT(g+yTR) >> 5) & 0x07) | (LIMIT(r+yTR) & 0xF8);

		/* Skip down to next line to write out bottom two pixels */
		rgb += 2 * rowPixels;

		rgb[0] = ((LIMIT(b+yBL) >> 3) & 0x1F) | ((LIMIT(g+yBL) << 3) & 0xE0);
		rgb[1] = ((LIMIT(g+yBL) >> 5) & 0x07) | (LIMIT(r+yBL) & 0xF8);

		rgb[2] = ((LIMIT(b+yBR) >> 3) & 0x1F) | ((LIMIT(g+yBR) << 3) & 0xE0);
		rgb[3] = ((LIMIT(g+yBR) >> 5) & 0x07) | (LIMIT(r+yBR) & 0xF8);
	}
}

/*
 * convert a YUV420P to a rgb image
 */
int v4l_yuv420p2rgb (unsigned char *rgb_out, unsigned char *yuv_in,
		int width, int height, int bits)
{
	const int numpix = width * height;
	const unsigned int bytes = bits >> 3;
	int h, w, y00, y01, y10, y11, u, v;
	unsigned char *pY = yuv_in;
	unsigned char *pU = pY + numpix;
	unsigned char *pV = pU + numpix / 4;
	unsigned char *pOut = rgb_out;

	if (!rgb_out || !yuv_in)
		return -1;

	for (h = 0; h <= height - 2; h += 2) {
		for (w = 0; w <= width - 2; w += 2) {
			y00 = *(pY);
			y01 = *(pY + 1);
			y10 = *(pY + width);
			y11 = *(pY + width + 1);
			u = (*pU++) - 128;
			v = (*pV++) - 128;

			v4l_copy_420_block (y00, y01, y10, y11, u, v, width, pOut, bits);
	
			pY += 2;
			pOut += bytes << 1;

		}
		pY += width;
		pOut += width * bytes;
	}
	return 0;
}

/* init v4l device for image capture */
int v4l_init(MyVID * myvid)
{
    int devfd;
    int i, ret = 0;
    int input;

    struct video_capability vid_caps;
    struct video_channel vid_chnl;
    struct video_picture vid_pict;
    struct video_window vid_win;
    struct video_mbuf vid_buf;

    /*
     * open the video4linux device 
     */
    int max_try = 5;
    while ((devfd = open(myvid->device, O_RDWR)) == -1 && --max_try)
	sleep(1);
    if (devfd == -1) {
	fprintf(stderr, "Can't open device %s\n", VIDEO_DEV);
	return -1;
    }
    myvid->devfd = devfd;

    /*
     * get capabilities of the device 
     */
    if (ioctl(devfd, VIDIOCGCAP, &vid_caps) == -1) {
	perror("ioctl (VIDIOCGCAP)");
	return -2;
    }

    myvid->maxinputs = vid_caps.channels;

    /* print channel names */
    if (verbose >= 2) {
	for (i = 0; i < myvid->maxinputs; i++) {
	    vid_chnl.channel = i;
	    if (ioctl(devfd, VIDIOCGCHAN, &vid_chnl) == -1) {
		perror("ioctl (VIDIOCGCHAN)");
	    }
	    printf("The name of input channel #%d is \"%s\".\n", i, vid_chnl.name);
	}
    }

    input = myvid->input;

    if (input == -1) {		// if no input channel specified, use first composite.
	input = (myvid->maxinputs == 1) ? 0 : 1;	// tv often use channel 0, skip it.
    }

    if (myvid->input > myvid->maxinputs - 1) {
	input = myvid->maxinputs - 1;
	printf("Warning: Input channel number over range. use %d instead.\n", input);
    }

    if (verbose >= 2) {
	printf("Total %d input channel(s). Using channel #%d.\n", myvid->maxinputs, input);
	printf("Supported resolution range: %dx%d => %dx%d\n\n", vid_caps.minwidth,
	       vid_caps.minheight, vid_caps.maxwidth, vid_caps.maxheight);
    }

    /*
     * get default video properties and then set new value 
     */
    if (ioctl(devfd, VIDIOCGPICT, &vid_pict) == -1) {
	perror("ioctl (VIDIOCGPICT)");
	ret = 1;
    }
    vid_pict.brightness = myvid->brightness;
    vid_pict.contrast = myvid->contrast;
    if (myvid->depth == 1){
	vid_pict.palette = VIDEO_PALETTE_GREY;
	myvid->palette = VIDEO_PALETTE_GREY;
    }
    else if (myvid->depth == 3){
	vid_pict.palette = VIDEO_PALETTE_RGB24;
	myvid->palette = VIDEO_PALETTE_RGB24;
    }
    else {
	printf("Wrong color depth. only 1 or 3 is valid.\n");
	return -3;
    }

    if (ioctl(devfd, VIDIOCSPICT, &vid_pict) != -1) {
	vid_pict.palette = VIDEO_PALETTE_YUV420P;
	myvid->palette = VIDEO_PALETTE_YUV420P;
        if (ioctl(devfd, VIDIOCSPICT, &vid_pict) == 0) {
	    if(verbose >= 1){
		fprintf(stderr, "Device NO RGB24 palette support, Using YUV420P instead.\n\n");
	    }
        }
	else{
	    perror("ioctl (VIDIOCSPICT)");
	    ret = 1;
	}
    }

    /*
     * get default and then set new channel and norm 
     */
    vid_chnl.channel = input;
    if (ioctl(devfd, VIDIOCGCHAN, &vid_chnl) == -1) {
	perror("ioctl (VIDIOCGCHAN)");
    }
    vid_chnl.channel = input;
    vid_chnl.norm = myvid->norm;
    if (ioctl(devfd, VIDIOCSCHAN, &vid_chnl) == -1) {
	perror("ioctl (VIDIOCSCHAN)");
	ret = 2;
    }

    /*
     * check if drvier support mmap mode and do init 
     */
    if (ioctl(devfd, VIDIOCGMBUF, &vid_buf) == -1) {
	fprintf(stderr, "No mmap support, Use read instead.\n");
	myvid->mmaped = 0;
	if (ioctl(devfd, VIDIOCGWIN, &vid_win) != -1) {
	    vid_win.width = myvid->width;
	    vid_win.height = myvid->height;
	    if (ioctl(devfd, VIDIOCSWIN, &vid_win) == -1) {
		perror("ioctl (VIDIOCSWIN)");
		return -2;
	    }
	}
    }
    else {
	myvid->mmaped = 1;
	myvid->mbufsize = vid_buf.size;
    }
    return ret;
}

/*
 * read rgb image from v4l device
 * return: mmap'ed buffer and size
 */
int get_image(MyVID * myvid)
{
    struct video_mmap vid_mmap;
    char *map, *convbuf;
    int len;
    int framesize = myvid->width * myvid->height * myvid->depth;

    if (myvid->mmaped) {
	map = mmap(0, myvid->mbufsize, PROT_READ | PROT_WRITE, MAP_SHARED, myvid->devfd, 0);
	if ((unsigned char *) -1 == (unsigned char *) map) {
	    perror("mmap()");
	    return -1;
	}

	vid_mmap.format = myvid->palette;
	vid_mmap.frame = 0;
	vid_mmap.width = myvid->width;
	vid_mmap.height = myvid->height;
	if (ioctl(myvid->devfd, VIDIOCMCAPTURE, &vid_mmap) == -1) {
	    perror("VIDIOCMCAPTURE");
	    munmap(map, myvid->mbufsize);
	    return -1;
	}
	if (ioctl(myvid->devfd, VIDIOCSYNC, &vid_mmap) == -1) {
	    perror("VIDIOCSYNC");
	    munmap(map, myvid->mbufsize);
	    return -1;
	}
	memcpy(myvid->buf, map, framesize);
	munmap(map, myvid->mbufsize);
    }
    else {
	len = read(myvid->devfd, myvid->buf, framesize);
	if (len <= 0) {
	    return -1;
	}
    }
    if(myvid->palette == VIDEO_PALETTE_YUV420P){
        convbuf = malloc(framesize);
        v4l_yuv420p2rgb(convbuf, myvid->buf, myvid->width, myvid->height, myvid->depth * 8);
        memcpy(myvid->buf, convbuf, framesize);
        free(convbuf);
    }
    return 0;
}

/*
 * Write image buffer to jpeg file.
 */
void put_image_jpeg(char *filename, IMG image, int quality)
{
#ifdef HAVE_JPEG
    char *buf = image.buf;
    int width = image.width;
    int height = image.height;
    int depth = image.depth;

    int y, x, line_width;
    JSAMPROW row_ptr[1];
    struct jpeg_compress_struct cjpeg;
    struct jpeg_error_mgr jerr;
    char *line;
    FILE *fd;

    if ((fd = fopen(filename, "w+")) == NULL) {
	perror("fopen JPG");
	exit(-2);
    }
    line = malloc(width * 3);
    if (!line)
	return;
    cjpeg.err = jpeg_std_error(&jerr);
    jpeg_create_compress(&cjpeg);
    cjpeg.image_width = width;
    cjpeg.image_height = height;
    cjpeg.input_components = 3;
    cjpeg.in_color_space = JCS_RGB;
    if (depth == 1) {
	cjpeg.input_components = 1;
	cjpeg.in_color_space = JCS_GRAYSCALE;
    }
    jpeg_set_defaults(&cjpeg);

    jpeg_set_quality(&cjpeg, quality, TRUE);
    cjpeg.dct_method = JDCT_FASTEST;
    jpeg_stdio_dest(&cjpeg, fd);

    jpeg_start_compress(&cjpeg, TRUE);

    row_ptr[0] = line;

    if (depth == 1) {
	line_width = width;
	for (y = 0; y < height; y++) {
	    row_ptr[0] = buf;
	    jpeg_write_scanlines(&cjpeg, row_ptr, 1);
	    buf += line_width;
	}
    }
    else {
	line_width = width * 3;
	for (y = 0; y < height; y++) {
	    for (x = 0; x < line_width; x += 3) {
		line[x] = buf[x + 2];
		line[x + 1] = buf[x + 1];
		line[x + 2] = buf[x];
	    }
	    jpeg_write_scanlines(&cjpeg, row_ptr, 1);
	    buf += line_width;
	}
    }
    jpeg_finish_compress(&cjpeg);
    jpeg_destroy_compress(&cjpeg);
    free(line);
    fclose(fd);
#else
    fprintf(stderr, "NOT Compiled with JPEG support!\n\n");
    exit(2);
#endif
}

/*
 * Write image buffer to png file.
 */
void put_image_png(char *filename, IMG image)
{
#ifdef HAVE_PNG
    int y;
    char *p;
    png_infop info_ptr;
    FILE *fd;
    int width = image.width;
    int height = image.height;
    int depth = image.depth;

    png_structp png_ptr;

    if ((fd = fopen(filename, "w+")) == NULL) {
	perror("fopen PNG");
	exit(-2);
    }

    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (!png_ptr)
	return;
    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
	return;

    png_init_io(png_ptr, fd);
    if (depth == 1) {
	png_set_IHDR(png_ptr, info_ptr, width, height,
		     8, PNG_COLOR_TYPE_GRAY, PNG_INTERLACE_NONE,
		     PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
    }
    else {
	png_set_IHDR(png_ptr, info_ptr, width, height,
		     8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
		     PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
    }
    png_set_bgr(png_ptr);
    png_write_info(png_ptr, info_ptr);
    p = image.buf;
    for (y = 0; y < height; y++) {
	png_write_row(png_ptr, p);
	p += width * depth;
    }
    png_write_end(png_ptr, info_ptr);
#else
    fprintf(stderr, "NOT Compiled with PNG support!\n\n");
    exit(2);
#endif
}

/* 
 * Create direcotry recursively
 */
int pmkdir(char *path)
{
    struct stat st;
    int umask_save;
    char *dir, *parent;

    dir = strdup(path);
    if (stat(dir, &st) < 0 && errno == ENOENT) {
	parent = dirname(dir);
	pmkdir(parent);
	if (verbose >= 1)
	    printf("Making dir: %s\n", path);
	umask_save = umask(0);
	umask(umask_save);
	if (mkdir(path, 00777 & ~umask_save) < 0) {
	    perror("MKDIR ");
	};
    }
    free(dir);
    return 0;
}

/*
 * Divide image into sx * sy blocks. For each splited block, 
 * calculate its average RGB color, then store these RGB color 
 * values into array r.
 */
int proc_img(RGB24 * r, IMG image, int sx, int sy)
{
    char *buf = image.buf;
    int w = image.width;
    int h = image.height;
    int depth = image.depth;

    unsigned char *p;
    int x, y, m, n;
    int blk_wid = w / sx, blk_hgt = h / sy;
    int x0, y0, x1, y1;
    int s = blk_wid * blk_hgt;
    unsigned int R, G, B;
    int r_offset, g_offset;

    if (depth == 1) {
	r_offset = 0;
	g_offset = 0;
    }
    else {
	r_offset = 2;
	g_offset = 1;
    }
    for (n = 1; n <= sy; n++) {
	for (m = 1; m <= sx; m++) {
	    x0 = (m - 1) * blk_wid;
	    y0 = (n - 1) * blk_hgt;
	    x1 = blk_wid * m - 1;
	    y1 = blk_hgt * n - 1;
	    R = G = B = 0;
	    for (y = y0; y <= y1; y++) {
		for (x = x0; x <= x1; x++) {
		    p = buf + (y * w + x) * depth;
		    R += *(p + r_offset);
		    G += *(p + g_offset);
		    B += *p;
		}
	    }
	    r->red = R / s;
	    r->green = G / s;
	    r->blue = B / s;
	    r++;
	}
    }
    return s;
}

/*
 * Compare each RGB element in array a and b (array size: num),
 * if grayscale, use red field only, set depth=1, else depth=3. 
 * if RGB color difference > c_limit, mark it as hot point. 
 * if count hot point > b_limit, return 1.
 */
int cmp_blks(RGB24 * a, RGB24 * b, int num, int depth, int c_limit, int b_limit)
{
    int i, count;
    int color_diff[8192];

    count = 0;
    for (i = 0; i < num; i++) {
	if (depth == 1) {
	    color_diff[i] = abs(a[i].red - b[i].red) * 3;
	}
	else {
	    color_diff[i] = abs(a[i].red - b[i].red) +
		abs(a[i].green - b[i].green) + abs(a[i].blue - b[i].blue);
	}
	if (color_diff[i] > c_limit) {
	    if (++count >= b_limit)
		return 1;
	}
    }
    return 0;
}

int cmp_blks2(RGB24 * a, RGB24 * b, int sx, int sy, int depth, int c_limit, int b_limit)
{
    int x, y, count, i = 0;
    int color_diff[8192];

    count = 0;
    for (y = 0; y < sy; y++) {
	for (x = 0; x < sx; x++) {
	    if (depth == 1) {
		color_diff[i] = abs(a[i].red - b[i].red) * 3;
	    }
	    else {
		color_diff[i] = abs(a[i].red - b[i].red) +
		    abs(a[i].green - b[i].green) + abs(a[i].blue - b[i].blue);
	    }
	    if (color_diff[i] > c_limit) {
		count++;
	    }
	    printf("%3d|", color_diff[i]);
	    i++;
	}
	printf("\n");
    }
    printf("Hot blocks: %d / %d\n\n", count, i);
    if (count >= b_limit)
	return 1;
    else
	return 0;
}

/*
 * main()
 */
int main(int argc, char *argv[])
{
    int c, n = 0;
    int ofmt = DEF_FMT;
    int create_dir = 1;
    int ptimestamp = PRINT_TIME;
    int prog_run = 0, ret, pstat;
    int blks, slow_count, comp_last, comp_older;
    unsigned int bufsize;
    char *buf;
    IMG image;
    RGB24 *colors_new, *colors_last, *colors_older, *ptr, textcolor;
    char *basedir = BASEDIR;
    char *userprog = NULL;
    pid_t pid = -1;
    char lasttime[32], str_date[32], str_time[32], dirname[255], filename[255];
    time_t now;
    struct tm *tm;
    float w = DEF_INTERVAL;
    unsigned int delay = w * 1000000;

    FONT font;
    MyVID myvid;

    char *device = VIDEO_DEV;
    int width = DEF_WIDTH, height = DEF_HEIGHT;
    int input = IN_DEFAULT;
    int norm = NORM_DEFAULT;
    int brightness = DEF_BRIGHTNESS;
    int contrast = DEF_CONTRAST;
    int depth = 3;
    int blk_x = DEF_BLK_X;
    int blk_y = DEF_BLK_Y;
    int color_diff_ts = DEF_COLOR_TS;	/* default color diff threshold */
    int hotblk_ts = blk_x * blk_y / 64 + 1;	/* default hot block threshold */

    int jpeg_quality = QUAL_DEFAULT;	/* default jpeg quality setting */

    if (signal(SIGINT, sig_quit) == SIG_ERR) {
	printf("can not setup SIGINT handler.\n");
	exit(1);
    }
    if (signal(SIGTSTP, sig_tstp) == SIG_ERR) {
	printf("can not setup SIGTSTP handler.\n");
	exit(1);
    }
    if (signal(SIGUSR1, sig_usr1) == SIG_ERR) {
	printf("can not setup SIGUSR1 handler.\n");
	exit(1);
    }

    while ((c = getopt(argc, argv, "B:C:O:a:b:c:d:e:s:q:f:n:i:l:w:Dhmovt")) != EOF) {
	/*
	 * B: brightness -B 32768
	 * C: contrast -C 32768
	 * D: run in daemon mode
	 * O: base directory for output
	 * o: don't create sub-directories
	 * a: horizontal split number x vertical split number -a 16x8
	 * b: threshold of hot block number (0-blk_x*blk_y)
	 * c: threshold of color diff for each block (0-255*3)
	 * d: v4l device name
	 * e: execute a external program while motion detected.
	 * s: width x height  -s 640x480
	 * q: jpeg quality -q 85
	 * f: output file format
	 * t: print timestamp on image
	 * m: grayscale mode
	 * n: input video norm
	 * i: input channel #
	 * l: logfile -l /var/log/mvc.log
	 * w: delay time for each capture(seconds), -w 0.5
	 * h: help
	 */
	switch (c) {
	case 'B':
	    sscanf(optarg, "%d", &brightness);
	    break;
	case 'C':
	    sscanf(optarg, "%d", &contrast);
	    break;
	case 'D':
	    daemon_flag = 1;
	    break;
	case 'O':
	    basedir = optarg;
	    break;
	case 'd':
	    device = optarg;
	    break;
	case 's':
	    sscanf(optarg, "%dx%d", &width, &height);
	    break;
	case 'a':
	    sscanf(optarg, "%dx%d", &blk_x, &blk_y);
	    hotblk_ts = blk_x * blk_y / 64 + 1;
	    break;
	case 'b':
	    sscanf(optarg, "%d", &hotblk_ts);
	    break;
	case 'c':
	    sscanf(optarg, "%d", &color_diff_ts);
	    break;
	case 'e':
	    userprog = optarg;
	    break;
	case 'f':
	    if (strcasecmp("jpeg", optarg) == 0) {
		ofmt = FMTJPEG;
	    }
	    else if (strcasecmp("png", optarg) == 0) {
		ofmt = FMTPNG;
	    }
	    break;
	case 't':
	    ptimestamp = 1;
	    break;
	case 'm':
	    depth = 1;
	    break;
	case 'n':
	    if (strcasecmp("ntsc", optarg) == 0) {
		norm = VIDEO_MODE_NTSC;
	    }
	    else if (strcasecmp("pal", optarg) == 0) {
		norm = VIDEO_MODE_PAL;
	    }
	    else if (strcasecmp("secam", optarg) == 0) {
		norm = VIDEO_MODE_SECAM;
	    }
	    else if (strcasecmp("auto", optarg) == 0) {
		norm = VIDEO_MODE_AUTO;
	    }
	    break;
	case 'o':
	    create_dir = 0;
	    break;
	case 'q':
	    sscanf(optarg, "%d", &jpeg_quality);
	    break;
	case 'w':
	    sscanf(optarg, "%f", &w);
	    delay = w * 1000000;
	    break;
	case 'i':
	    sscanf(optarg, "%d", &input);
	    break;
	case 'l':
	    logfile = optarg;
	    break;
	case 'v':
	    verbose++;
	    break;
	case 'h':
	default:
	    usage();
	    exit(1);
	}
    }

    printf("MVC version " VERSION "\n\n");

    if (verbose >= 2) {
	fprintf(stderr, "--------------- Debug Info (level %d)---------------\n", verbose);
	fprintf(stderr, " ImageSize:     %dx%d\n", width, height);
	fprintf(stderr, " ImageFormat:   %s\n", (ofmt == FMTPNG) ? "PNG" : "JPEG");
	fprintf(stderr, " JpegQuality:   %d\n", jpeg_quality);
	fprintf(stderr, " VideoDEV:      %s\n", device);
	fprintf(stderr, " VideoInput(-1 for Default): %d\n", input);
	fprintf(stderr, " VideoNORM(PAL=0, NTSC=1, SECAM=2, AUTO=3): %d\n", norm);
	fprintf(stderr, " BlockDivision: %dx%d\n", blk_x, blk_y);
	fprintf(stderr, " ColorDiff_TS:  %d\n", color_diff_ts);
	fprintf(stderr, " HotBlock_TS:   %d\n", hotblk_ts);
	fprintf(stderr, " Color Mode:    %s\n", (depth == 1) ? "GRAYSCALE" : "COLOR");
	fprintf(stderr, " Interval:      %0.2f second\n", w);
	fprintf(stderr, " Daemon Log:    %s\n", logfile);
	fprintf(stderr, " Base DIR:      %s\n", basedir);
	fprintf(stderr, "---------------------------------------------------\n\n");
    }

    if (daemon_flag) {
	daemonizenow();
    }
    else {
	printf("Press CTRL-Z to switch program to daemon mode.\n\n");
    }

    myvid.device = device;
    myvid.width = width;
    myvid.height = height;
    myvid.depth = depth;
    myvid.input = input;
    myvid.norm = norm;
    myvid.brightness = brightness;
    myvid.contrast = contrast;

    block_sigs();		/* bttv driver ioctl hates signal */
    if (v4l_init(&myvid) != 0) {
	fprintf(stderr, "v4l init fault!\n");
	exit(0);
    }
    unblock_sigs();

    bufsize = myvid.width * myvid.height * myvid.depth;
    myvid.buf = malloc(bufsize);

    blks = blk_x * blk_y;
    colors_new = calloc(blks, sizeof(RGB24));
    colors_last = calloc(blks, sizeof(RGB24));
    colors_older = calloc(blks, sizeof(RGB24));

    /* grub a init image as reference */
    if (verbose >= 3) {
	printf("Initializing first image...\n");
    }
    block_sigs();
    image.buf = myvid.buf;
    image.width = width;
    image.height = height;
    image.depth = depth;
    if (get_image(&myvid) == 0) {
	unblock_sigs();
	proc_img(colors_last, image, blk_x, blk_y);
	memcpy(colors_older, colors_last, blks * sizeof(RGB24));
    }

    /* begin loop */
    slow_count = 0;
    while (!quitflag) {
	if (verbose >= 3) {
	    printf("Doing capture and compare...\n");
	}
	block_sigs();
	if (get_image(&myvid) == 0) {
	    unblock_sigs();
	    buf = myvid.buf;
	    image.buf = myvid.buf;
	    proc_img(colors_new, image, blk_x, blk_y);
	    if (verbose >= 4) {
		printf("Result:\n");
		comp_last = cmp_blks2(colors_new, colors_last, blk_x, blk_y, depth, color_diff_ts,
				      hotblk_ts);
	    }
	    else {
		comp_last = cmp_blks(colors_new, colors_last, blks, depth, color_diff_ts, hotblk_ts);
	    }
	    comp_older = 0;
	    if (slow_count) {
		comp_older = cmp_blks(colors_new, colors_older, blks, depth,
				      color_diff_ts, hotblk_ts);
	    }

	    if (comp_last || comp_older || usr1_flag) {	/* motion detected or found specified flag to do actions */
		usr1_flag = 0;

		/* Write Image */
		time(&now);
		tm = localtime(&now);

		if (verbose >= 3 && !comp_last && comp_older) {
		    printf("slow motion detected. slow_count=%d\n", slow_count);
		}

		if (create_dir) {
		    strftime(str_time, 31, "%H.%M.%S", tm);
		    strftime(str_date, 31, "%Y-%m-%d", tm);
		    sprintf(dirname, "%s/%s/%02d", basedir, str_date, tm->tm_hour);
		    pmkdir(dirname);

		    if (strcmp(str_time, lasttime) == 0)
			n++;
		    else
			n = 0;
		    strcpy(lasttime, str_time);
		    sprintf(filename, "%s/img-%s-%d", dirname, str_time, n);
		}
		else {
		    strftime(str_time, 31, "%Y%m%d-%H%M%S", tm);
		    if (strcmp(str_time, lasttime) == 0)
			n++;
		    else
			n = 0;
		    strcpy(lasttime, str_time);
		    sprintf(filename, "%s/img-%s-%d", basedir, str_time, n);
		}

		if (ptimestamp) {
		    strftime(str_time, 20, "%Y/%m/%d %H:%M:%S", tm);
		    textcolor.red = 0;
		    textcolor.green = 0;
		    textcolor.blue = 0;
		    font=FONTSIZE;
		    gputs(image, width - font.char_width*strlen(str_time)-2, height - font.char_height - 1, str_time, textcolor, font);
		    textcolor.green = 255;
		    gputs(image, width - font.char_width*strlen(str_time)-3, height - font.char_height - 2, str_time, textcolor, font);
		}

		if (ofmt == FMTJPEG) {
		    strcat(filename, ".jpg");
		    put_image_jpeg(filename, image, jpeg_quality);
		}
		else if (ofmt == FMTPNG) {
		    strcat(filename, ".png");
		    put_image_png(filename, image);
		}
		if (verbose >= 1) {
		    printf("Wrote: %s\a\n", filename);
		}

		/* fork and execute external program if needed */
		if (userprog != NULL) {
		    if (!prog_run) {
			prog_run = 1;
			if (verbose >= 3) {
			    printf("Start user program: %s\n", userprog);
			}
			if ((pid = fork()) == -1) {
			    perror("Fork: ");
			    exit(-1);
			}
		    }
		    if (pid == 0) {	// child
			ret = system(userprog);
			exit(ret);
		    }
		    else {	// parent
			if (waitpid(-1, &pstat, WNOHANG) != 0) {
			    if (WIFEXITED(pstat) && verbose >= 3) {
				printf("User program is done. exit code %d.\n", WEXITSTATUS(pstat));
			    }
			    prog_run = 0;
			}
		    }
		}
		/* update 'color_older' array */
		memcpy(colors_older, colors_new, blks * sizeof(RGB24));
		slow_count = 0;
	    }
	    else {		/* no motion detected */
		slow_count++;
	    }

	    if (slow_count >= SLOW_COUNT_MAX) {
		memcpy(colors_older, colors_new, blks * sizeof(RGB24));
		slow_count = 0;
	    }
	    ptr = colors_last;
	    colors_last = colors_new;
	    colors_new = ptr;
	}
	else {
	    fprintf(stderr, "Error: Can't get image\n");
	}

	if (tstp_flag) {
	    daemonizenow();
	    daemon_flag = 1;
	    tstp_flag = 0;
	    verbose = 1;
	}
	usleep(delay);
    }				/* end of while */
    exit(0);
}

